from enum import Enum


class QueryType(Enum):
    PLURAL = 'PLURAL'
    SINGULAR = 'SINGULAR'


class MutationType(Enum):
    CREATE = 'CREATE'
    UPDATE = 'UPDATE'
    UPSERT = 'UPSERT'
    DELETE = 'DELETE'
    LOCAL = 'LOCAL'  # assume local object exists in DB


'''
children(['room', 'inventories']) # ! DEFAULTS TO SAME MUTATION (CREATE, UPDATE)
children(
  room = MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
  inventories = MutationType.CREATE, # ! MODE, NO CHILDREN IMPLICIT
)
children({
  'room': MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
  'inventories': {
    '__mode': MutationType.CREATE, # ! MODE, CHILDREN EXPLICIT
    'tank': MutationType.UPSERT
  },
})

1. CreateProgram
2. Link program.id to every plan
3. CreateDemand (with prev. program.id)
4. CreateOrders
  4.1 Create plans with related order.id
'''


class MutationMode(Enum):
    BATCH = 'BATCH'
    NESTED = 'NESTED'



from datetime import datetime
from tzlocal import get_localzone
from dateutil import tz
from dateutil.parser import parse


# Save some global variables here
utc = tz.UTC
LOCAL_TIMEZONE = get_localzone()


def get_utc_date(date):
    if not date.tzinfo:
        local_date = date.replace(tzinfo=LOCAL_TIMEZONE)
    else:
        local_date = date
    utc_date = local_date.astimezone(utc)
    return utc_date


def get_local_date(date):
    if not date.tzinfo:  # if false, already localized, skip
        local_date = date.replace(tzinfo=LOCAL_TIMEZONE)
    else:
        local_date = date
    return local_date


def iso8601_to_local_date(iso_str):
    if not type(iso_str) == str:
        return iso_str  # do nothing if not string, expect it to be a date already
    utc_date = datetime.strptime(iso_str, "%Y-%m-%dT%H:%M:%SZ")
    utc_date = utc_date.replace(tzinfo=utc)
    return utc_date.astimezone(LOCAL_TIMEZONE)


def get_iso8601_str(date):
    return date.strftime("%Y-%m-%dT%H:%M:%SZ")


def get_datetime_object(datetime, format_type):
    """This function returns a datetime object.

    Args:
        datetime (str): Datetime string
        format_type (str): datetime format

    Returns:
        object: Datetime object
    """
    datetime = parse(datetime)

    if format_type == "Time":
        return datetime.time()

    elif format_type == "Date":
        return datetime.date()

    else:
        return datetime



import os
from typing import Union, List, TextIO
from gstorm.helpers.Logging import log, LogLevel
from pydash import camel_case


def inmemory_cleanup(src):
    src_path = os.path.join(os.getcwd(), src)
    output = []
    with open(src_path, 'r') as src_file:
        for line in src_file:
            is_whitespace = not line.strip()
            if is_whitespace:
                output.append(line)
                continue
            line = line.rstrip()  # clean
            line = line.split('#')[0].rstrip()  # remove comment
            # append if not empty
            if line:
                output.append(line + '\n')
        output.append('\n')
    return output


def scalar_map(field_typename):
    return {
        'Boolean': 'Boolean',
        'Integer': 'Int',
        'Float': 'Float',
        'String': 'String',
        'Text': 'Text',
        'Json': 'map',
        'Jsonb': 'Jsonb',
        'Date': 'Date',
        'Time': 'Time',
        'Datetime': 'DateTime',
        'ID': 'ID'
    }[field_typename]


def get_field_kind(field_typename):
    scalar_types = [
        'Boolean',
        'Integer',
        'Float',
        'String',
        'Text',
        'Json',
        'Jsonb',
        'Date',
        'Time',
        'Datetime',
        'ID'
    ]
    if field_typename in scalar_types:
        return 'SCALAR'
    else:
        return 'OBJECT'


'''
Syntactic analizer for a graphql schema

things to consider:

# ! 1. general structure of a type
type TYPE_NAME @opts {
  ...
}
# ! 2. general structure of an enum
enum ENUM_TYPE @opts {
  ...
}
# ! 3. All basic types
type MyType {
  enabled: Boolean
  quantity: Int
  weight: Float
  name: String
  description: Text
  config: Json
  shipAt: Datetime
}
# ! 4. Complex type relations
# ? 4.1 ComplexType1.ListOf.ComplexType2 (Cannot be a list of BasicType)
# ? 4.2 ComplexType1.hasOne.ComplexType2 (Cannot be a hasOne of BasicType)
# ? 4.3 ComplexType1.ComplexType2 (ComplexType2 Must exist)
type User {
  name: String!
  address: Address @has_one
  posts: [Blogpost]
}

type Address {
  street: String
  city: String
  number: Integer
}

type Blogpost {
  author: User
  title: String @unique
}

type UserSkill {
  user: User
  skill: Skill
  level: Integer!
}

type Skill {
  name: String
}
'''


def string_bool_eval(str_bool: Union[str, bool]) -> Union[str, bool]:
    '''
    Function that takes a string or a bool, if a string is given, if the string
    is TRUE or FALSE when upper is used, returns the boolean equivalent,
    if not, returns the string
    if a boolean is passed, returns the boolean value passed

    Args
    --------
    str_bool: Union[str, bool], String or bool to be evaluated

    Returns
    --------
    Union[str, bool], bool if input is bool or if is a boolean in string form
                      else, string
    '''
    evaluation = str_bool
    if isinstance(str_bool, str):
        upper = str_bool.upper()
        match upper:
            case "FALSE":
                evaluation = False
            case "TRUE":
                evaluation = True
    return evaluation


def gql_schema_parse(schema_file_lines: Union[TextIO, List[str]]) -> dict:
    '''
    Function that takes a file or a list of strings and generates a schema

    Args
    -------
    schema_file_lines: Union[TextIO, List[str]], An open text file or a list
                       of strings

    Returns
    -------
    dict, dict that has the schema found in the input
    '''
    # known states: TYPE_SEARCH, PARSING_TYPE, PARSING_ENUM, PARSING_INDEX
    state = 'TYPE_SEARCH'
    schema = {
        'types': {},
        'enums': {},
        'indexes': {}
    }  # dict of types, enums and indexes
    known_labels = ['public', "opt-out"]
    # context
    current_type = None
    current_field = None
    for originalLine in schema_file_lines:
        line = originalLine.strip()
        tokens = line.split()
        if (state == 'TYPE_SEARCH'):
            current_type = {'labels': {
                label: False for label in known_labels}}
            for i, token in enumerate(tokens):
                if tokens[i] == 'type':
                    state = 'PARSING_TYPE'
                    current_type['kind'] = 'TYPE'
                    current_type['fields'] = []
                    continue
                elif token.startswith("@"):
                    label = token.split('@')
                    if label[-1] not in known_labels:
                        raise ValueError(
                            f"The type label {label[-1]} is not a known label, check: {originalLine}")
                    current_type['labels'][label[1]] = True
                    continue
                elif tokens[i] == 'enum':
                    state = 'PARSING_ENUM'
                    current_type['kind'] = 'ENUM'
                    current_type['enumValues'] = []
                elif tokens[i] == 'index':
                    state = 'PARSING_INDEX'
                    current_type['kind'] = 'INDEX'
                    current_type['indexes'] = []
                elif token != "{":
                    current_type['name'] = token
                    continue
                continue
        elif (state == 'PARSING_TYPE'):
            if len(tokens) == 0:
                # blank line inside the type
                continue
            elif tokens[0] == '}':
                # end of type definition
                state = 'TYPE_SEARCH'
                schema['types'][current_type['name']] = current_type
                current_type = None
            else:
                # type body (should be a field definition)
                # known labels are camelCased (hasOne and has_one are the same)
                known_labels_fields = [
                    'hasOne', 'unique', 'default', "hasManyThrough", "hasMany", "public", "optOut"]
                field_is_list = False
                field_typename = None
                current_field = {label: False for label in known_labels_fields}
                current_field['default'] = None
                for i, token in enumerate(tokens):
                    if ":" in token and i == 0:
                        current_field['name'] = token.strip(':')
                    elif "!" in token and token[-1] == '!':
                        current_field['is_required'] = True
                        field_typename = token.strip(
                            '!').strip('[]').strip('!')
                    elif token.startswith("@"):
                        label = camel_case(token.strip('@'))
                        if "id" in label:
                            continue
                        if not any([knownLabel in label for knownLabel in known_labels_fields]):
                            raise ValueError(
                                f"The field label '{token}' is not a known label, check: {originalLine}")
                        label = [
                            knownLabel for knownLabel in known_labels_fields if knownLabel in label][0]
                        current_field[label] = True
                        if 'default' in label:
                            default = tokens[i + 1][:-1].strip('"').strip("'")
                            default_proccesed = string_bool_eval(default)
                            current_field[label] = default_proccesed
                        if 'hasManyThrough' in label:
                            current_field[label] = tokens[i+1][:-1]
                    else:
                        field_typename = token.strip(
                            '!').strip('[]').strip('!') if not field_typename else field_typename
                        if '[' in token and ']' in token:
                            field_is_list = True
                    if i < len(tokens)-1:
                        continue
                    field_kind = get_field_kind(field_typename)
                    current_field['type'] = {
                        'name': field_typename if field_kind == 'OBJECT' else scalar_map(field_typename),
                        'kind': 'LIST' if field_is_list else field_kind
                    }
                    current_type['fields'].append(current_field)
        elif (state == 'PARSING_ENUM'):
            if len(tokens) == 0:
                # blank line inside the enum
                continue
            elif tokens[0] == '}':
                # end of enum definition
                state = 'TYPE_SEARCH'
                # we put the enum in both types and enums, so we can check if a type is an enum
                schema['types'][current_type['name']] = current_type
                schema['enums'][current_type['name']] = current_type
                current_type = None
            else:
                # enum body (should be a EnumValue definition)
                current_type['enumValues'].append(tokens[0].strip())
        elif (state == 'PARSING_INDEX'):
            if len(tokens) == 0:
                # blank line inside the index
                continue
            elif tokens[0] == '}':
                # end of index definition
                state = 'TYPE_SEARCH'
                schema['indexes'][current_type['name']] = current_type
                current_type = None
            else:
                # index body (should be a IndexValue definition)
                current_type['indexes'].append(
                    {tokens[0].strip(':'): tokens[1]})
    # we now know every schema type, we need to update the fields type kind to ENUM if needed:
    for typename, item in schema['types'].items():
        if item['kind'] == 'ENUM':
            continue
        for index, field in enumerate(item['fields']):
            if not field['type']['name'] in schema['types']:
                # may be an error, but leave it to the validation function
                continue
            if (
                field['type']['kind'] == 'OBJECT'
                and schema['types'][field['type']['name']]['kind'] == 'ENUM'
            ):
                schema['types'][typename]['fields'][index]['type']['kind'] = 'ENUM'
    return schema


def load_schema_from_file(src):
    try:
        clean_schema_lines = inmemory_cleanup(src)
        return gql_schema_parse(clean_schema_lines)
    except FileNotFoundError:
        log(LogLevel.ERROR, f'File {src} not found')
        return


def prettify_query(query):
    tabs = 0
    pretty_query = ''
    parsing_string = False
    inside_brackets = False
    for i, ch in enumerate(query):
        # get context
        prev_ch = query[i-1] if i > 0 else ''
        next_ch = query[i+1] if i < len(query) - 1 else ''
        if not inside_brackets and ch == '{':
            inside_brackets = True
        elif not inside_brackets:
            pretty_query += ch
            continue
        if not parsing_string and ch == '"':
            parsing_string = True
            pretty_query += ch
            continue
        if parsing_string and ch != '"':
            pretty_query += ch
            continue
        if parsing_string and ch == '"':
            parsing_string = False
            pretty_query += ch
            continue
        tabs += 1 if ch in ['{', '('] else (-1 if next_ch in ['}', ')'] else 0)
        if ch != ' ' or prev_ch == ':':
            pretty_query += ch
            continue
        if next_ch in ['{', '(']:
            pretty_query += ch  # add whitespace normally
        elif next_ch in ['}', ')']:
            pretty_query += f"\n{'  ' * tabs}"
        else:
            pretty_query += f"\n{'  ' * tabs}"
    return pretty_query



from .gql import setup_gql
from .gql_schema_helpers import load_schema_from_file
from .os_helpers import (
    silent_removefile,
    query_yes_no
)

from .str_handling import (
    capitalize,
    objPathToTypeParam
)




# ! Referencia para posible pre-procesador de queries

# SOURCE
query = '''
query MyQuery(
  $name: String!
  $capacity: Integer = 1000
){
  gov_tanks: tanks(
    name: $name
    capacity: $capacity
    type: "Gobierno"
  ){
    id
    name
    insertedAt
    location: room {
      id
      name
    }
    inventories (orderBy: { asc: ID } limit: 1) {
      id
      volume
    }
  }
  rooms {
    id
    name
  }
}
'''

# RESULT
parsed_query = {
    '__type': 'Query',
    'label': 'MyQuery',
    'variables': {
        'name': {
            '__type': 'String',
            'required': True,
            'default': None
        },
        'capacity': {
            '__type': 'Integer',
            'required': False,
            'default': 1000
        }
    },
    'operations': [{
        'label': 'gov_tanks',
        'name': 'tanks',
        'params': [
            'name': GraphQLVariable('name')
            'capacity': GraphQLVariable('capacity')
            'type': "Gobierno"
        ],
        'fields': [{
            'label': None,
            'name': 'id',
            'params': [],
            'fields': []
        }, {
            'label': None,
            'name': 'name',
            'params': [],
            'fields': []
        }, {
            'label': None,
            'name': 'insertedAt',
            'params': [],
            'fields': []
        }, {
            'label': 'location',
            'name': 'room',
            'params': [],
            'fields': [{
                'label': None,
                'name': 'id',
                'params': [],
                'fields': []
            }, {
                'label': None,
                'name': 'name',
                'params': [],
                'fields': []
            }]
        }, {
            'label': None,
            'name': 'inventories',
            'params': [
                'orderBy': {'asc': 'ID'},
                'limit': 1
            ],
            'fields': [{
                'label': None,
                'name': 'id',
                'params': [],
                'fields': []
            }, {
                'label': None,
                'name': 'volume',
                'params': [],
                'fields': []
            }]
        }]
    }, {
        'label': None,
        'name': 'rooms',
        'params': [],
        'fields': [{
            'label': None,
            'name': 'id',
            'params': [],
            'fields': []
        }, {
            'label': None,
            'name': 'name',
            'params': [],
            'fields': []
        }]
    }]
}



from __future__ import annotations
from typing import Union, List
from gstorm.enums import MutationMode, MutationType
import pydash as _pd
import gstorm


def create(
    data: Union[gstorm.GraphQLType, List[gstorm.GraphQLType]],
    mode: MutationMode = MutationMode.BATCH
) -> gstorm.MutationBuilder:
    """Builds instance of MutationBuilder to sync GraphQLType objects created locally with the remote DB.

    Parameters
    ----------
    data : Union[gstorm.GraphQLType, List[gstorm.GraphQLType]]
      GraphQLType instance(s) to create in the DB
    mode : MutationMode, optional
      allows to upload data by different upload mechanisms, by default MutationMode.BATCH

    Returns
    -------
    MutationBuilder
      Instance of MutationBuilder class, responsible of building mutation string and communication with DB.
    """
    return gstorm.MutationBuilder(data=data, mode=mode)


def save_multi_create(
    to_be_published: List[GraphQLType],
    children_list: List[str],
    api_limit: int = 100
):
    '''Gstorm 'create' method with slicing for batch-size-limited API.\n
    E. g.:\n
    # Where plan1 & plant2 are instances of a GraphQL-schema-based class (e. g. BbtPlan)\n
    plans = [plan1, plan2]
    errors = multi_create(plans, ['program', 'brightBeer', 'tank'])\n
    # Equivalent to:
    gstorm.create(plans).children(['program', 'brightBeer', 'tank']).run()
    # But sliced...
    '''
    # Since each element amounts for len(children_list),
    # the maximum size an upload batch may have is api_limit//to_be_published
    max_batch_size = api_limit if not children_list else api_limit//len(
        children_list)

    # We need a variable to report any errors when posting items.
    response = {'errors': [], 'messages': []}

    chunked = _pd.chunk(to_be_published, max_batch_size)

    # Iterate over the chunked guy:
    for chunk in chunked:

        # Post the batch
        result = create(chunk).children(children_list).run()

        # Errors, if any, are to be reported
        response['errors'].extend([s for s in result['successful'] if not s])
        response['messages'].extend([s for s in result['messages']])

    # Report errors
    return response


def update(data: Union[gstorm.GraphQLType, List[gstorm.GraphQLType]],
           mode: MutationMode = MutationMode.BATCH,
           findByKeys: List[str] = None
           ) -> gstorm.MutationBuilder:
    """Builds instance of MutationBuilder to sync GraphQLType objects created locally with the
       remote DB.

    Parameters
    ----------
    data : Union[gstorm.GraphQLType, List[gstorm.GraphQLType]]
      GraphQLType instance(s) to update in the DB
    mode : MutationMode, optional
      allows to upload data by different upload mechanisms, by default MutationMode.BATCH
    keys: List[str], optonal
      Allows the user to set keys to be used for the findby method within the database
    Returns
    -------
    MutationBuilder
      Instance of MutationBuilder class, responsible of building mutation string and
      communication with DB.
    """
    return gstorm.MutationBuilder(data=data, mode=mode, Type=MutationType.UPDATE, findByKeys=findByKeys)


def upsert(data: Union[gstorm.GraphQLType, List[gstorm.GraphQLType]],
           mode: MutationMode = MutationMode.BATCH,
           findByKeys: List[str] = None) -> gstorm.MutationBuilder:
    """Builds instance of MutationBuilder to sync GraphQLType objects created locally with the
       remote DB.

    Parameters
    ----------
    data : Union[gstorm.GraphQLType, List[gstorm.GraphQLType]]
      GraphQLType instance(s) to update in the DB
    mode : MutationMode, optional
      allows to upload data by different upload mechanisms, by default MutationMode.BATCH
    keys: List[str], optonal
      Allows the user to set keys to be used for the findby method within the database
    Returns
    -------
    MutationBuilder
      Instance of MutationBuilder class, responsible of building mutation string and
      communication with DB.
    """
    return gstorm.MutationBuilder(data=data, mode=mode, Type=MutationType.UPSERT, findByKeys=findByKeys)


def delete():
    raise NotImplementedError('gstorm.delete not implemented')



one_to_one_map = {
    "ID": ("Union[str, int]", "0"),
    "String": ("str", "''"),
    "Int": ("int", "0"),
    "Float": ("float", "0.0"),
    "Date": ("date", "date.today()"),
    "Time": ("time", "datetime.now().time()"),
    "DateTime": ("datetime", "datetime.now()"),
    "Boolean": ("bool", "False"),
    "map": ("dict", "{}"),
    "Json": ("dict", "{}"),
    "Jsonb": ("dict", "{}"),
    "Text": ("str", "''"),
    "Enum": ("str", "''")
}


def default_map(var_type, variable):
    """This function determines the 'python values'

    Args:
        var_type (str): Attribut type
        variable (str): Default value

    Returns:
        tuple: returns a tuple with the class data
    """
    if var_type == 'Int':
        return ("int", int(variable))
    elif var_type == 'Float':
        return ("float", float(variable))
    elif var_type == 'Date':
        return ('date', f'get_datetime_object({variable}, "{var_type}")')
    elif var_type == 'Time':
        return ('time', f'get_datetime_object({variable}, "{var_type}")')
    elif var_type == 'DateTime':
        return ('datetime', f'get_datetime_object({variable}, "{var_type}")')
    elif var_type == "Boolean":
        return ("bool", variable)
    elif var_type == "map":
        return ("dict", variable)
    elif var_type == "String":
        return ("str", f"'{variable}'")
    else:  # Enum
        return (var_type, variable)


imports = """
from __future__ import annotations
from typing import List, Union
from datetime import datetime, date, time
from gstorm.GraphQLType import GraphQLType
from gstorm.helpers.typing_helpers import convert_to, list_convert_to, gql_repr, gql_list_repr
from gstorm.helpers.date_helpers import iso8601_to_local_date, get_datetime_object
import attr
from enum import Enum
"""

example = {'Author': {'fields': [{'has_one': False,
                                  'name': 'name',
                                  'type': {'kind': 'SCALAR', 'name': 'String'},
                                  'unique': True},
                                 {'has_one': False,
                                  'name': 'lastName',
                                  'type': {'kind': 'SCALAR', 'name': 'String'},
                                  'unique': True},
                                 {'has_one': False,
                                  'name': 'finishTurn',
                                  'type': {'kind': 'SCALAR', 'name': 'Time'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'dateOfBirth',
                                  'type': {'kind': 'SCALAR', 'name': 'Date'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'active',
                                  'type': {'kind': 'SCALAR', 'name': 'Boolean'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'blogPosts',
                                  'type': {'kind': 'LIST', 'name': 'BlogPost'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'comments',
                                  'type': {'kind': 'LIST', 'name': 'Comment'},
                                  'unique': False},
                                 {'has_one': True,
                                  'name': 'avatar',
                                  'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'categories',
                                  'type': {'kind': 'LIST', 'name': 'Category'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'session',
                                  'type': {'kind': 'SCALAR', 'name': 'Float'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'email',
                                  'type': {'kind': 'SCALAR', 'name': 'String'},
                                  'unique': False}],
                      'kind': 'TYPE',
                      'name': 'Author'},
           'Avatar': {'fields': [{'has_one': False,
                                  'name': 'imageUrl',
                                  'type': {'kind': 'SCALAR', 'name': 'String'},
                                  'unique': True},
                                 {'has_one': False,
                                  'name': 'avatarConfigurations',
                                  'type': {'kind': 'LIST', 'name': 'AvatarConfiguration'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'comments',
                                  'type': {'kind': 'LIST', 'name': 'Comment'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'owner',
                                  'type': {'kind': 'OBJECT', 'name': 'Author'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'blogPost',
                                  'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'providers',
                                  'type': {'kind': 'LIST', 'name': 'Reply'},
                                  'unique': False},
                                 {'has_one': False,
                                  'name': 'consumers',
                                  'type': {'kind': 'LIST', 'name': 'Reply'},
                                  'unique': False}],
                      'kind': 'TYPE',
                      'name': 'Avatar'},
           'AvatarConfiguration': {'fields': [{'has_one': False,
                                               'name': 'guid',
                                               'type': {'kind': 'SCALAR',
                                                        'name': 'String'},
                                               'unique': True},
                                              {'has_one': False,
                                               'name': 'config',
                                               'type': {'kind': 'OBJECT',
                                                        'name': 'Avatar'},
                                               'unique': False},
                                              {'has_one': False,
                                               'name': 'version',
                                               'type': {'kind': 'SCALAR',
                                                        'name': 'String'},
                                               'unique': False}],
                                   'kind': 'TYPE',
                                   'name': 'AvatarConfiguration'},
           'BlogPost': {'fields': [{'has_one': False,
                                    'name': 'author',
                                    'type': {'kind': 'OBJECT', 'name': 'Author'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'name',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': True},
                                   {'has_one': False,
                                    'name': 'body',
                                    'type': {'kind': 'SCALAR', 'name': 'Text'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'title',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'lastName',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'dateOfPublication',
                                    'type': {'kind': 'SCALAR', 'name': 'Date'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'statusWord',
                                    'type': {'kind': 'ENUM', 'name': 'StatusWord'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'comments',
                                    'type': {'kind': 'LIST', 'name': 'Comment'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'approvedAt',
                                    'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                                    'unique': False},
                                   {'has_one': True,
                                    'name': 'alias',
                                    'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'tag',
                                    'type': {'kind': 'ENUM', 'name': 'Tag'},
                                    'unique': False},
                                   {'has_one': True,
                                    'name': 'wordCount',
                                    'type': {'kind': 'OBJECT', 'name': 'WordCount'},
                                    'unique': False}],
                        'kind': 'TYPE',
                        'name': 'BlogPost'},
           'Category': {'fields': [{'has_one': False,
                                    'name': 'name',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': True},
                                   {'has_one': False,
                                    'name': 'description',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'createdBy',
                                    'type': {'kind': 'OBJECT', 'name': 'Author'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'editedBy',
                                    'type': {'kind': 'OBJECT', 'name': 'Author'},
                                    'unique': False}],
                        'kind': 'TYPE',
                        'name': 'Category'},
           'Comment': {'fields': [{'has_one': False,
                                   'name': 'author',
                                   'type': {'kind': 'OBJECT', 'name': 'Author'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'supremeAuthor',
                                   'type': {'kind': 'OBJECT', 'name': 'Author'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'blogPost',
                                   'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'body',
                                   'type': {'kind': 'SCALAR', 'name': 'String'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'recievedAt',
                                   'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'reportedAt',
                                   'type': {'kind': 'SCALAR', 'name': 'DateTime'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'avatar',
                                   'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'styles',
                                   'type': {'kind': 'LIST', 'name': 'Style'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'finishedAt',
                                   'type': {'kind': 'SCALAR', 'name': 'Int'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'type',
                                   'type': {'kind': 'SCALAR', 'name': 'Int'},
                                   'unique': False},
                                  {'has_one': False,
                                   'name': 'tag',
                                   'type': {'kind': 'ENUM', 'name': 'Tag'},
                                   'unique': False}],
                       'kind': 'TYPE',
                       'name': 'Comment'},
           'Employee': {'fields': [{'has_one': False,
                                    'name': 'name',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'email',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': True},
                                   {'has_one': False,
                                    'name': 'department',
                                    'type': {'kind': 'SCALAR', 'name': 'String'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'employer',
                                    'type': {'kind': 'OBJECT', 'name': 'Employee'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'employees',
                                    'type': {'kind': 'LIST', 'name': 'Employee'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'boss',
                                    'type': {'kind': 'OBJECT', 'name': 'Employee'},
                                    'unique': False},
                                   {'has_one': False,
                                    'name': 'workers',
                                    'type': {'kind': 'LIST', 'name': 'Employee'},
                                    'unique': False}],
                        'kind': 'TYPE',
                        'name': 'Employee'},
           'Equipment': {'fields': [{'has_one': False,
                                     'name': 'name',
                                     'type': {'kind': 'SCALAR', 'name': 'String'},
                                     'unique': True},
                                    {'has_one': False,
                                     'name': 'description',
                                     'type': {'kind': 'SCALAR', 'name': 'Text'},
                                     'unique': False},
                                    {'has_one': False,
                                     'name': 'information',
                                     'type': {'kind': 'SCALAR', 'name': 'map'},
                                     'unique': False}],
                         'kind': 'TYPE',
                         'name': 'Equipment'},
           'OrderComment': {'fields': [{'has_one': False,
                                        'name': 'description',
                                        'type': {'kind': 'SCALAR', 'name': 'Text'},
                                        'unique': False}],
                            'kind': 'TYPE',
                            'name': 'OrderComment'},
           'Reply': {'fields': [{'has_one': False,
                                 'name': 'title',
                                 'type': {'kind': 'SCALAR', 'name': 'String'},
                                 'unique': True},
                                {'has_one': False,
                                 'name': 'provider',
                                 'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                                 'unique': False},
                                {'has_one': False,
                                 'name': 'consumer',
                                 'type': {'kind': 'OBJECT', 'name': 'Avatar'},
                                 'unique': False},
                                {'has_one': False,
                                 'name': 'struct',
                                 'type': {'kind': 'SCALAR', 'name': 'map'},
                                 'unique': False}],
                     'kind': 'TYPE',
                     'name': 'Reply'},
           'StatusWord': {'enumValues': ['APPROVED', 'PENDING', 'REJECTED'],
                          'kind': 'ENUM',
                          'name': 'StatusWord'},
           'Style': {'fields': [{'has_one': False,
                                 'name': 'description',
                                 'type': {'kind': 'SCALAR', 'name': 'Text'},
                                 'unique': False},
                                {'has_one': False,
                                 'name': 'comment',
                                 'type': {'kind': 'OBJECT', 'name': 'Comment'},
                                 'unique': False}],
                     'kind': 'TYPE',
                     'name': 'Style'},
           'Tag': {'enumValues': ['QUESTION', 'DEPENDENCY', 'NEWS'],
                   'kind': 'ENUM',
                   'name': 'Tag'},
           'WordCount': {'fields': [{'has_one': False,
                                     'name': 'outputBlogPost',
                                     'type': {'kind': 'OBJECT', 'name': 'BlogPost'},
                                     'unique': False},
                                    {'has_one': False,
                                     'name': 'count',
                                     'type': {'kind': 'SCALAR', 'name': 'Int'},
                                     'unique': False}],
                         'kind': 'TYPE',
                         'name': 'WordCount'}}



import sys
from functools import lru_cache
from typing import Type, Dict, Any, ForwardRef, _eval_type
from importlib import import_module
from enum import Enum
from gstorm import BaseGraphQLType


def is_enum(x):
    return Enum in type(x).__mro__


def is_gql_subtype(x):
    return BaseGraphQLType in type(x).__mro__


def enum_cast(x):
    return x.value if is_enum(x) else x


def convert_to(models_path, classname):
    """Returns a casting function to determined type
    Parameters
    ----------
    models_path : str
      String denoting the path to the models directory: "path.to.models"
    classname : str
      Model name (Ej: "Author", "User", etc)
    Returns
    -------
    function
      function to cast to the specified type
    Raises
    ------
    Exception
      When the class was not found in the available models
    """
    models = import_module(models_path)

    def concrete_cast(value):
        if hasattr(models, classname):
            cls = getattr(models, classname)
        else:
            raise Exception(
                f'"{classname}" model not found inside models module ({models_path})')
        '''casts to specified type'''
        if value is None:
            return None
        if type(value) == dict:
            return cls(**value)
        elif type(value) == str or is_enum(value):
            return cls(value)
        elif is_gql_subtype(value):
            return value
        else:
            return value
    return concrete_cast


def list_convert_to(models_path: str, classname: str):
    """Returns a casting function to determined type

    Parameters
    ----------
    models_path : str
      module path where to look for GraphQL models (ej: "tests.models")
    classname : str
      concrete type to convert-to

    Returns
    -------
    function
        function to cast to the specified type
    """
    models = import_module(models_path)

    def concrete_cast(values):
        if hasattr(models, classname):
            cls = getattr(models, classname)
        else:
            raise Exception(
                f'"{classname}" model not found inside models module ({models_path})')
        '''casts list items to specified type'''
        casted_values = []
        for value in values:
            if type(value) == dict:
                casted_values.append(cls(**value))
            elif type(value) == cls:
                init_values = {
                    k: v for k, v in value.__dict__.items() if not k.startswith('_')}
                casted_values.append(cls(**init_values))
        return casted_values
    return concrete_cast


def gql_repr(value):
    if value is None:
        return 'None'
    return f'{type(value).__name__}(...)'


def gql_list_repr(values):
    if values is None:
        return 'None'
    if len(values) == 0:
        return '[]'
    else:
        return f'[{type(values[0]).__name__}({len(values)})]'


@lru_cache
def attrs_with_hints(model: Type) -> Dict[str, Any]:
    """Method to return the attributes of a given class
    with the corresponding hints.

    Args:
    ------
        - model (Type): Model to obtain its attributes with hints

    Return:
    ------
        - Dict with the attr as key and the hint as value
    """
    mapped_attrs: Dict[str, Any] = {}  # Dictionary for the attributes
    for base_class in model.__mro__:
        # Iterate over the annotations of the base class
        for key, hint in base_class.__dict__.get("__annotations__", {}).items():
            # Obtain the import from the file where it resides this base class
            global_variables = sys.modules[base_class.__module__].__dict__
            if hint in global_variables:
                mapped_attrs[key] = global_variables[hint]
            elif hint in global_variables["__builtins__"]:
                mapped_attrs[key] = global_variables["__builtins__"][hint]
            else:
                # For those that are not directly in the global variables, try to evaluate them
                # as a forward reference. For Python <=3.8, it would not fail the `try` block and
                # it would work as expected. In Python >=3.8, if it fail with a `NameError`, then
                # it's an GraphQLType model.
                try:
                    mapped_attrs[key] = _eval_type(
                        ForwardRef(hint), global_variables, locals())
                except NameError:
                    mapped_attrs[key] = ForwardRef(hint)
    return mapped_attrs



import attr


@attr.s
class GraphQLEnum():
    """Wraps an enumeration string to identify it as an enum
    May be needed to identify string literals as enums in GraphQLBuilder.
    Example:
    "L101" -> str
    "INSERTED_AT" -> Enum
    "SUCCESS" -> Enum
    "Gobierno" -> str
    etc
    """
    value: str = attr.ib()
